//=====================================================================================
//
// Project: 
// Constant-time bilateral filter
// 
// Authors: 
// Daniel Sage, Kunal N. Chaudhury
// Biomedical Imaging Group (BIG)
// Ecole Polytechnique Federale de Lausanne (EPFL)
// Lausanne, Switzerland
//
// Date:
// 4 February 2011
//
// Reference:
// Kunal N. Chaudhury, Daniel Sage, and Michael Unser,
// Fast O(1) bilateral filtering using trigonometric range kernels
// IEEE Transactions on Image Processing, submitted Dec. 2010
//
// Conditions of use:
// You'll be free to use this software for research purposes, but you
// should not redistribute it without our consent. In addition, we 
// expect you to include a citation or acknowledgment whenever 
// you present or publish results that are based on it.
//
//=====================================================================================

import ij.IJ;
import ij.ImagePlus;
import ij.WindowManager;
import ij.gui.GUI;
import ij.process.ByteProcessor;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;

import additionaluserinterface.GridPanel;
import additionaluserinterface.SpinnerDouble;
import additionaluserinterface.WalkBar;

public class Bilateral_Filter_Higher_Dimensional extends JDialog implements WindowListener, ActionListener, Runnable {

	private Thread thread = null;
	private JButton bnRun = new JButton("Run");
	private SpinnerDouble spnRange = new SpinnerDouble(5, 0, 10000.0, 1);
	private SpinnerDouble spnSpace = new SpinnerDouble(2, 0, 10000.0, 1);
	private WalkBar walk = new WalkBar("(c) 2011 EPFL, BIG", false, false, true);
	
	/**
	*/
	public Bilateral_Filter_Higher_Dimensional() {
		super(new JFrame(), "Bilateral Filter Higher Dimensional");
		
		GridPanel panel = new GridPanel();
		panel.place(0, 0, new JLabel("Sigma Range"));
		panel.place(0, 1, spnRange);
		panel.place(1, 0, new JLabel("Sigma Space"));
		panel.place(1, 1, spnSpace);
		panel.place(3, 1, bnRun);
		
		bnRun.addActionListener(this);
		
		GridPanel pn = new GridPanel(8);
		pn.place(0, 0, panel);
		pn.place(1, 0, walk);
		
		walk.getButtonClose().addActionListener(this);
		
		addWindowListener(this);
		getContentPane().add(pn);
		pack();
		setResizable(false);
		GUI.center(this);
		setVisible(true);
	}
	
	/**
	 * Implements the run for the Runnable.
	 */
	public synchronized  void actionPerformed(ActionEvent e) {

		Object source = e.getSource();
    	if (source == walk.getButtonClose()) {
      		dispose();
    	}
		
		else if (source == bnRun) {
			if (thread == null) {
				thread = new Thread(this);
				thread.setPriority(Thread.MIN_PRIORITY);
				thread.start();
			}
		}
	}
		
	/**
	 * Implements the run for the Runnable.
	 */
	public void run() {
		ImagePlus imp = WindowManager.getCurrentImage();
		if (imp == null) {
			IJ.error("No open image.");
			thread = null;
			return;
		}
		if (imp.getStackSize() > 1) {
			IJ.error("Only process 2D image.");
			thread = null;
			return;
		}
		if (imp.getType() != ImagePlus.GRAY8) {
			IJ.error("Only process 8-bits image.");
			thread = null;
			return;
		}
		bnRun.setEnabled(false);
		process(imp);
		bnRun.setEnabled(true);
		thread = null;
	}
	
	/**
	*/
	private void process(ImagePlus imp) {
		
		int nx = imp.getWidth();
		int ny = imp.getHeight();
		int nr = 256;
		
		walk.reset();
		float[][][] vol = new float[nx][ny][nr];
		walk.progress("Allocation", 10);
		
		ByteProcessor bp = (ByteProcessor)imp.getProcessor();
		for(int j=0; j<ny; j++)
		for(int i=0; i<ny; i++) {
			int k = (int)bp.getPixelValue(i, j);
			vol[i][j][k] = 1.0f;
		}
		walk.progress("Build map", 30);
		
		double ss = spnSpace.get();
		double sr = spnRange.get();
		
		gaussian(vol, nx, ny, nr, ss, ss, sr);
		
		ByteProcessor outMax = new ByteProcessor(nx, ny);
		int kmax = 0;
		for(int j=0; j<ny; j++)
		for(int i=0; i<ny; i++) {
			float max = -Float.MAX_VALUE;
			for(int k=0; k<nr; k++) {
				if (vol[i][j][k] > max) {
					max = vol[i][j][k];
					kmax = k;
				}
			}
			outMax.putPixelValue(i, j, kmax);
		}
		
		(new ImagePlus("Output Max", outMax)).show();
		walk.finish();
	}

	/**
	 * Implements the methods for the WindowListener.
	 */
	public void windowActivated(WindowEvent e) 		{}
	public void windowClosed(WindowEvent e) 		{}
	public void windowDeactivated(WindowEvent e) 	{}
	public void windowDeiconified(WindowEvent e)	{}
	public void windowIconified(WindowEvent e)		{}
	public void windowOpened(WindowEvent e)			{}			
	public void windowClosing(WindowEvent e) 		{
		dispose(); 
	}
	
	/**
	*/
	public void gaussian(float[][][] vol, int nx, int ny, int nz, double sigmaX, double sigmaY, double sigmaZ) {
		int n = 3;
		float N = (float)n;
		float poles[] = new float[n];

		if (nx > 1 && sigmaX > 0.0) {
			float s2 = (float)(sigmaX * sigmaX);
			float alpha = 1.0f + (N/s2) - (float)(Math.sqrt(N*N+2*N*s2)/s2);
			poles[0] = poles[1] = poles[2] = alpha;
			float line[] = new float[nx];
			for (int z=0; z<nz; z++)
			for (int y=0; y<ny; y++) {
				for(int x=0; x<nx; x++)
					line[x] = vol[x][y][z];
				line = convolveIIR(line, poles);
				for(int x=0; x<nx; x++)
					vol[x][y][z] = line[x];
			}
		}
		walk.progress("Smoothing X", 50);

		if (ny > 1 && sigmaY > 0.0) {
			float s2 = (float)(sigmaY * sigmaY);
			float alpha = 1.0f + (N/s2) - (float)(Math.sqrt(N*N+2*N*s2)/s2);
			poles[0] = poles[1] = poles[2] = alpha;
			float line[] = new float[ny];
			for (int x=0; x<nx; x++)
			for (int z=0; z<nz; z++) {
				for(int y=0; y<ny; y++)
					line[y] = vol[x][y][z];
				line = convolveIIR(line, poles);
				for(int y=0; y<ny; y++)
					vol[x][y][z] = line[y];
			}
		}
		walk.progress("Smoothing Y", 70);

		if (nz > 1 && sigmaZ > 0.0) {
			float s2 = (float)(sigmaZ * sigmaZ);
			float alpha = 1.0f + (N/s2) - (float)(Math.sqrt(N*N+2*N*s2)/s2);
			poles[0] = poles[1] = poles[2] = alpha;
			float line[] = new float[nz];
			for (int y=0; y<ny; y++)
			for (int x=0; x<nx; x++) {
				for(int z=0; z<nz; z++)
					line[z] = vol[x][y][z];
				line = convolveIIR(line, poles);
				for(int z=0; z<nz; z++)
					vol[x][y][z] = line[z];
			}
		}
		walk.progress("Smoothing Z", 90);
	}

	
	/**
	* Convolve with with a Infinite Impluse Response filter (IIR)
	*
	* @param input  	1D input signal
	* @param poles    	1D array containing the poles of the filter
	*/
	public static float[] convolveIIR(float[] input, float poles[]) {
		float lambda = 1.0f;
		int l = input.length;
		float[] output = new float[l];
		for (int k = 0; k < poles.length; k++) {
			lambda = lambda * (1.0f - poles[k]) * (1.0f - 1.0f / poles[k]);
		}
		for (int n = 0; n < l; n++) {
			output[n] = input[n] * lambda;
		}
		for (int k = 0; k < poles.length; k++) {
			output[0] = getInitialCausalCoefficientMirror(output, poles[k]);
			for (int n = 1; n < l; n++) {
				output[n] = output[n] + poles[k] * output[n - 1];
			}
			output[l-1] = getInitialAntiCausalCoefficientMirror(output, poles[k]);
			for (int n = l - 2; 0 <= n; n--) {
				output[n] = poles[k] * (output[n+1] - output[n]);
			}
		}
		return output;
	}

	/**
	 */
	private static float getInitialAntiCausalCoefficientMirror(float[] c, float z) {
		return((z * c[c.length - 2] + c[c.length - 1]) * z / (z * z - 1.0f));
	}

	/**
	 */
	private static float getInitialCausalCoefficientMirror(float[] c, float z) {
		float tolerance = 10e-6f;
		float z1 = z;
		float zn = (float)Math.pow(z, c.length - 1);
		float sum = c[0] + zn * c[c.length - 1];
		int horizon = c.length;

		if (0.0 < tolerance) {
			horizon = 2 + (int)(Math.log(tolerance) / Math.log(Math.abs(z)));
			horizon = (horizon < c.length) ? (horizon) : (c.length);
		}
		zn = zn * zn;
		for (int n = 1; (n < (horizon - 1)); n++) {
			zn = zn / z;
			sum = sum + (z1 + zn) * c[n];
			z1 = z1 * z;
		}
		return(sum / (1.0f - (float)Math.pow(z, 2 * c.length - 2)));
	}

}
