//=====================================================================================
//
// Project: 
// Constant-time bilateral filter
// 
// Authors: 
// Daniel Sage, Kunal N. Chaudhury
// Biomedical Imaging Group (BIG)
// Ecole Polytechnique Federale de Lausanne (EPFL)
// Lausanne, Switzerland
//
// Date:
// 4 February 2011
//
// Reference:
// Kunal N. Chaudhury, Daniel Sage, and Michael Unser,
// Fast O(1) bilateral filtering using trigonometric range kernels
// IEEE Transactions on Image Processing, submitted Dec. 2010
//
// Conditions of use:
// You'll be free to use this software for research purposes, but you
// should not redistribute it without our consent. In addition, we 
// expect you to include a citation or acknowledgment whenever 
// you present or publish results that are based on it.
//
//=====================================================================================

import ij.IJ;
import ij.ImagePlus;
import ij.WindowManager;
import ij.gui.GUI;
import ij.process.ByteProcessor;
import ij.process.FloatProcessor;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;

import additionaluserinterface.GridPanel;
import additionaluserinterface.SpinnerDouble;
import additionaluserinterface.SpinnerInteger;
import additionaluserinterface.WalkBar;

public class Bilateral_Filter_Naive extends JDialog implements WindowListener, ActionListener, Runnable {

	private WalkBar walk = new WalkBar("(c) 2009 EPFL, Biomedical Imaging Group", false, false, true);
	private Thread thread = null;
	private JButton bnRun = new JButton("Run");
	private SpinnerDouble spnRange = new SpinnerDouble(50, 0, 10000.0, 1);
	private SpinnerDouble spnSpace = new SpinnerDouble(3, 0, 10000.0, 1);
	private SpinnerInteger spnTruncation = new SpinnerInteger(3, 1, 10, 1);
	private JCheckBox chkMultithread = new JCheckBox("Mutithread (row by row)", true);
	
	private double[][] gs;	// 2D discrete gaussian for space
	private double[] gr;	// 1D discrete gaussian for range
	private int ngs; 
	private int hgr;	// half-size of the range gaussian
	private ImagePlus imp;
	private FloatProcessor out;
	
	/**
	*/
	public Bilateral_Filter_Naive() {
		super(new JFrame(), "Bilateral Filter Naive");
		doDialog();
	}
	
	/**
	 * Build the dialog box.
	 */
	private void doDialog() {
		
		GridPanel panel = new GridPanel();
		panel.place(0, 0, new JLabel("Sigma Range"));
		panel.place(0, 1, spnRange);
		panel.place(1, 0, new JLabel("Sigma Space"));
		panel.place(1, 1, spnSpace);
		panel.place(2, 0, new JLabel("Gaussian Trunction (x sigma)"));
		panel.place(2, 1, spnTruncation);
		panel.place(3, 0, chkMultithread);
		panel.place(4, 1, bnRun);
		
		bnRun.addActionListener(this);
		
		GridPanel pn = new GridPanel(8);
		pn.place(0, 0, panel);
		pn.place(1, 0, walk);
		
		walk.getButtonClose().addActionListener(this);
		
		addWindowListener(this);
		getContentPane().add(pn);
		pack();
		setResizable(false);
		GUI.center(this);
		setVisible(true);
	}
	
	/**
	 * Implements the run for the Runnable.
	 */
	public synchronized  void actionPerformed(ActionEvent e) {

		Object source = e.getSource();
    	if (source == walk.getButtonClose()) {
      		dispose();
    	}
		
		else if (source == bnRun) {
			if (thread == null) {
				thread = new Thread(this);
				thread.setPriority(Thread.MIN_PRIORITY);
				thread.start();
			}
		}
	}
		
	/**
	 * Implements the run for the Runnable.
	 */
	public void run() {
		imp = WindowManager.getCurrentImage();
		
		if (imp == null) {
			IJ.error("No open image.");
			thread = null;
			return;
		}
		if (imp.getStackSize() > 1) {
			IJ.error("Only process 2D image.");
			thread = null;
			return;
		}
		if (imp.getType() != ImagePlus.GRAY8) {
			IJ.error("Only process 8-bits image.");
			thread = null;
			return;
		}
		bnRun.setEnabled(false);
		walk.reset();
		process();
		walk.finish();
		bnRun.setEnabled(true);
		thread = null;
	}
	
	/**
	*/
	private void process() {
		
		walk.reset();
		
		int ny = imp.getHeight();
		
		int trunc = spnTruncation.get();
		gs = generateDiscreteGaussian(spnSpace.get(), spnSpace.get(), trunc);
		gr = generateDiscreteGaussian(spnRange.get(), trunc);

		ngs = gs.length;
		hgr = gr.length / 2;

		walk.reset();
		if (chkMultithread.isSelected()) {
			try {
				for(int y=0; y<ny; y++) {
					FilterRow filter = new FilterRow(y);
					filter.start();
					filter.join();
				}
			}
			catch(InterruptedException e) {
				IJ.log("Error in the thread management");
				return;
			
			}
		}
		else {
			for(int y=0; y<ny; y++) {
				FilterRow filter = new FilterRow(y);
				filter.run();
			}
		}

		walk.finish();
		(new ImagePlus("BF Naive of " + imp.getTitle() + " " + spnSpace.get() + "-" + spnSpace.get() + "-"+ trunc, out)).show();
	}

	private double[][] generateDiscreteGaussian(double sigmax, double sigmay, int trunc) {
		int nx = 2*(int)Math.ceil(trunc*sigmax)+1;
		int ny = 2*(int)Math.ceil(trunc*sigmay)+1;
		int hx = nx/2;
		int hy = ny/2;
		double g[][] = new double[nx][ny];
		double s = 1.0 / (sigmax*sigmay*2);
		double cst = 1.0 / (2*Math.PI*sigmax*sigmay);
		for (int x=0; x<nx; x++)
		for (int y=0; y<ny; y++)
			g[x][y] = cst*Math.exp(-((x-hx)*(x-hx)+(y-hy)*(y-hy))*s);
		return g;
	}

	private double[] generateDiscreteGaussian(double sigma, int trunc) {
		int n = 2*(int)Math.ceil(trunc*sigma)+1;
		int h = n/2;
		double g[] = new double[n];
		double s = 1.0 / (sigma*sigma*2);
		double cst = 1.0 / (Math.sqrt(2*Math.PI)*sigma);
		for (int k=0; k<n; k++)
			g[k] = cst*Math.exp(-(k-h)*(k-h)*s);
		return g;
	}

	/**
	 * Implements the methods for the WindowListener.
	 */
	public void windowActivated(WindowEvent e) 		{}
	public void windowClosed(WindowEvent e) 		{}
	public void windowDeactivated(WindowEvent e) 	{}
	public void windowDeiconified(WindowEvent e)	{}
	public void windowIconified(WindowEvent e)		{}
	public void windowOpened(WindowEvent e)			{}			
	public void windowClosing(WindowEvent e) 		{
		dispose(); 
	}

	private class FilterRow extends Thread {
		
		private int y;
		
		public FilterRow(int y) {
			this.y = y;
			setPriority(Thread.MIN_PRIORITY);
		}
		
		public void run() {
			double q, gb;
			int nx = imp.getWidth();
			int ny = imp.getHeight();
			int hgs = ngs /2;
			ByteProcessor bp = (ByteProcessor)imp.getProcessor();
			byte[] pix = (byte[])bp.getPixels();
			
			float[][] block = new float[ngs][ngs];
			for(int x=0; x<nx; x++) {
				double value = 0;
				double norm = 0;
				//double p = image.getPixel(x, y, z);
				getNeighborhood(pix, x, y, nx, ny, block);
				double p = block[hgs][hgs];
				for (int i=0; i<ngs; i++)
				for (int j=0; j<ngs; j++) {
					q = block[i][j];
					int diffrange = (int)Math.round(Math.abs(p-q));
					if (diffrange < hgr) {
						gb = gs[i][j] * gr[diffrange+hgr];
						value += q * gb;
						norm += gb;
					}
				}
				out.putPixelValue(x, y, value/norm);
			}
		}
		
		/**
		*/
		private void getNeighborhood(byte pix[], int x, int y, int nx, int ny, float[][] buffer) {
			int xperiod = (nx <= 1 ? 1: 2*nx - 2);
			int yperiod = (ny <= 1 ? 1: 2*ny - 2);
			
			int xp, yp;
			int leni = buffer.length;
			int lenj = buffer[0].length;
			
			int xs = x - leni/2;
			int ys = y - lenj/2;
		
			for (int j=0; j<lenj; j++) {
				yp = ys + j;
				while (yp < 0)
					yp += yperiod;
				while (yp >= ny) {
					yp = yperiod - yp;
					yp = (yp < 0 ? -yp : yp);
				}
				yp *= nx;
				for (int i=0; i<leni; i++) {
					xp = xs + i;
					while (xp < 0)
						xp += xperiod;
					while (xp >= nx) {
						xp = xperiod - xp;
						xp = (xp < 0 ? -xp : xp);
					}
					buffer[i][j] = (float)(pix[xp + yp] & 0xFF);
				}
			}
		}
		
	}
	

}
