//=====================================================================================
//
// Project: 
// Constant-time bilateral filter
// 
// Authors: 
// Daniel Sage, Kunal N. Chaudhury
// Biomedical Imaging Group (BIG)
// Ecole Polytechnique Federale de Lausanne (EPFL)
// Lausanne, Switzerland
//
// Date:
// 4 February 2011
//
// Reference:
// Kunal N. Chaudhury, Daniel Sage, and Michael Unser,
// Fast O(1) bilateral filtering using trigonometric range kernels
// IEEE Transactions on Image Processing, submitted Dec. 2010
//
// Conditions of use:
// You'll be free to use this software for research purposes, but you
// should not redistribute it without our consent. In addition, we 
// expect you to include a citation or acknowledgment whenever 
// you present or publish results that are based on it.
//
//=====================================================================================

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.GenericDialog;
import ij.plugin.filter.PlugInFilter;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;

public class Bilateral_Filter_Polynomial implements PlugInFilter {

	private double radius;					// sigma Space
	private double sigmar;					// Sigma Range
	private boolean multithread = true;		// Run the gaussian in a multi-threaded fashion
	private boolean log = true;				// Display log messages
	private boolean saturate = true;		// Display log messages
	private ImagePlus imp;					// Input ImagePlus 
	private ImageStack stack;				// Output ImageStack
	private double chrono;
	
	/**
	* Implements the setup() method of the PlugInFilter.
	*/
	public int setup(String arg, ImagePlus imp) {
		if (imp == null) {
			IJ.error("No open image.");
			return DONE;
		}
		this.imp = imp;
		GenericDialog dlg = new GenericDialog("Bilateral Filter");
		dlg.addNumericField("Radius:", 10.0 , 1);
		dlg.addNumericField("Sigma Range:", 3.0 , 1);
		dlg.addMessage("N=4");
		dlg.addCheckbox("Saturate to the original dynamic range", true);
		dlg.addCheckbox("Multi-threaded Gaussian", true);
		dlg.addCheckbox("Display log", true);
					
		dlg.showDialog();
		if (dlg.wasCanceled())
			return DONE;
		
		radius = dlg.getNextNumber();
		sigmar = dlg.getNextNumber();
		saturate = dlg.getNextBoolean();
		multithread = dlg.getNextBoolean();
		log = dlg.getNextBoolean();
		
		int nx = imp.getWidth();
		int ny = imp.getHeight();
		stack = new ImageStack(nx, ny);
		if (log)
			IJ.log("");

		return(DOES_8G | DOES_16 | DOES_32 | DOES_STACKS);
	}
	
	/**
	* Implements the run() method of the PlugInFilter.
	*/
	public void run(ImageProcessor ip) {
		
		double chronoTotal = System.currentTimeMillis();		
		chrono = System.currentTimeMillis();		
		log("Bilateral filter (radius: " + radius +", sigma Range: " + sigmar + ")");
		
		// --------------------------------------------------------------------------
		// Convert the input image
		// --------------------------------------------------------------------------
		int nx = imp.getWidth();
		int ny = imp.getHeight();
		int size = nx*ny;
		double in[] = new double[size];
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		if (imp.getType() == ImagePlus.GRAY32) {
			for(int x=0; x<size; x++) {
				in[x] = (double)ip.getf(x);
				if (in[x] < min)
					min = in[x];
				if (in[x] > max)
					max = in[x];
			}
		}
		else if (imp.getType() == ImagePlus.GRAY16 || imp.getType() == ImagePlus.GRAY8) {
			for(int x=0; x<size; x++) {
				in[x] = (double)ip.get(x);
				if (in[x] < min)
					min = in[x];
				if (in[x] > max)
					max = in[x];
			}
		}

		log("Convert the input image (" + nx +"," + ny + ")");
		
		
		// --------------------------------------------------------------------------
		// Memory allocation
		// --------------------------------------------------------------------------
		log("Memory allocation ( 6 x double array of " + (size/1024) +" Kb )");		
		
		double c0 = 1.0;
		double c1 = -1.0 / (2.0 * sigmar * sigmar);
		double c2 = 1.0 / (8.0 * sigmar * sigmar * sigmar * sigmar);
		
		double[] d0  = new double[size];
		double[] d1  = new double[size];
		double[] d2  = new double[size];
		double[] d3  = new double[size];
		double[] d4  = new double[size];

		double[] fi0  = new double[size];
		double[] fi1  = new double[size];
		double[] fi2  = new double[size];
		double[] fi3  = new double[size];
		double[] fi4  = new double[size];
		double[] fi5  = new double[size];

		float[] po = new float[size];
		for(int x=0; x<size; x++) {
			double f1 = in[x];
			double f2 = f1*f1;
			double f3 = f2*f1;
			double f4 = f3*f1;
			
			fi0[x] = 1.0;
			fi1[x] = f1;
			fi2[x] = f2;
			fi3[x] = f3;
			fi4[x] = f4;
			fi5[x] = f4 * f1;
			
			d0[x]  = c0 + c1*f2 + c2*f4;	
			d1[x]  = - (2.0*c1*f1 + 4.0*c2*f3);	
			d2[x]  = c1 + 6.0*c2*f2;	
			d3[x]  = -4.0*c2*f1;	
			d4[x]  = c2;	
		}
		
		log("Initialisation");		
			
		// --------------------------------------------------------------------------
		// Gaussian
		// --------------------------------------------------------------------------
		int nGaussians = 6;
		Gaussian gaussians[] = new Gaussian[nGaussians];
		gaussians[0] = new Gaussian(fi0, radius, nx, ny);
		gaussians[1] = new Gaussian(fi1, radius, nx, ny);
		gaussians[2] = new Gaussian(fi2, radius, nx, ny);
		gaussians[3] = new Gaussian(fi3, radius, nx, ny);
		gaussians[4] = new Gaussian(fi4, radius, nx, ny);
		gaussians[5] = new Gaussian(fi5, radius, nx, ny);

		if (multithread) {
			try {
				for(int i=0; i<nGaussians; i++) {
					gaussians[i].start();
					gaussians[i].join();
				}
			}
			catch(InterruptedException e) {
				IJ.log("Error in the thread management");
				return;
			}
		}
		else {
			for(int i=0; i<nGaussians; i++)
				gaussians[i].run();
		}
		log("Gaussian convolutions");		
		
		// --------------------------------------------------------------------------
		// Final summation
		// --------------------------------------------------------------------------
		for(int x=0; x<size; x++) {
			double num = d0[x]*fi1[x] + d1[x]*fi2[x] + d2[x]*fi3[x] + d3[x]*fi4[x] + d4[x]*fi5[x];
			double dem = d0[x]*fi0[x] + d1[x]*fi1[x] + d2[x]*fi2[x] + d3[x]*fi3[x] + d4[x]*fi4[x];
			po[x] = (float)(num/dem); 
		}
		if (saturate) {
			float maxf = (float)max;
			float minf = (float)min;
			for(int x=0; x<size; x++) 
				po[x] = (po[x] < minf ? minf : (po[x] > maxf ? maxf : po[x]));
		}
		log("Final summation");		
		
		// --------------------------------------------------------------------------
		// Final summation
		// --------------------------------------------------------------------------
		FloatProcessor fp = new FloatProcessor(nx, ny, po, null);
		stack.addSlice("", fp);
		log("Display the result");		

		if (stack.getSize() == imp.getStackSize())
			(new ImagePlus("Bilateral Filter (" + radius + ") of " +imp.getTitle() , stack)).show();

		if (log)
			IJ.log("Time: " +(System.currentTimeMillis()-chronoTotal));
	}

	/** 
	 * Compute binomial coefficient.
	 */
	private double binomial(int n, int k) {
		if (k == 0)
			return 1;
		if (k == n)
			return 1;
		double facto = 1;
		for(int t=0; t<k; t++) 
			facto = facto * (n-t)/(k-t);
		return facto;
	}
	
	/** 
	 * Show a 1D float array as an image.
	 */
	private void show(double[] pix, String title, int nx, int ny) {
		float f[] = new float[nx*ny];
		for (int k=0; k<ny*nx; k++)
			f[k] = (float)pix[k];
		FloatProcessor fp = new FloatProcessor(nx, ny, f, null);
		(new ImagePlus(title, fp)).show();
	}
	
	/**
	* Print a log message + elapsed time.
	*/
	private void log(String message) {	
		 if (!log)
			return;
		IJ.log(message + " : " + (System.currentTimeMillis() - chrono) + " ms");		
		chrono = System.currentTimeMillis();
	}
	
	/**
	* Gaussian class.
	* Implementation of the Gaussian filter as a cascade of 3 exponential fitlers. 
	* The boundary conditions are mirroring.
	* Thread or directly by calling the run()
	*/
	public class Gaussian extends Thread {
	
		public double signal[];
		private double sigma;
		private int nx;
		private int ny;
		
		/**
		* Constructor based on the signal.
		*/
		public Gaussian(double signal[], double sigma, int nx, int ny) {
			this.signal = signal;
			this.sigma = sigma;
			this.nx = nx;
			this.ny = ny;
			setPriority(Thread.MIN_PRIORITY);
		}
		
		/**
		* Run method.
		*/
		public void run() {
			double s2 = sigma * sigma;
			double pole = 1.0 + (3.0/s2) - (Math.sqrt(9.0+6.0*s2)/s2);
		
			double row[]  = new double[nx];
			for (int y=0; y<ny*nx; y+=nx) {
				System.arraycopy(signal, y, row, 0, nx);
				row = convolveIIR_TriplePole(row, pole);
				System.arraycopy(row, 0, signal, y, nx);
			}

			double col[]  = new double[ny];
			for (int x=0; x<nx; x++) {
				for(int y=0; y<ny; y++)
					col[y] = signal[y*nx+x];
				col = convolveIIR_TriplePole(col, pole);
				for(int y=0; y<ny; y++)
					signal[y*nx+x] = col[y];
			}
		}
	
		/**
		* Convolve with with a Infinite Impulse Response filter (IIR)
		*/
		private double[] convolveIIR_TriplePole(double[] signal, double pole) {
			int l = signal.length;
			
			double lambda = 1.0;
			double[] output = new double[l];
			for (int k=0; k<3; k++) {
				lambda = lambda * (1.0 - pole) * (1.0 - 1.0 / pole);
			}
			for (int n=0; n<l; n++) {
				output[n] = signal[n] * lambda;
			}
			for (int k=0; k<3; k++) {
				output[0] = getInitialCausalCoefficientMirror(output, pole);
				for (int n=1; n<l ; n++) {
					output[n] = output[n] + pole * output[n - 1];
				}
				output[l-1] = getInitialAntiCausalCoefficientMirror(output, pole);
				for (int n=l-2; 0 <= n; n--) {
					output[n] = pole * (output[n+1] - output[n]);
				}
			}
			return output;
		}

		/**
		* Initial conditions
		*/
		private double getInitialAntiCausalCoefficientMirror(double[] c, double z) {
			return((z * c[c.length - 2] + c[c.length - 1]) * z / (z * z - 1.0));
		}
	
		/**
		* Initial conditions
		*/
		private double getInitialCausalCoefficientMirror(double[] c, double z) {
			double tolerance = 10e-6;
			double z1 = z, zn = Math.pow(z, c.length - 1);
			double sum = c[0] + zn * c[c.length - 1];
			int horizon = c.length;

			if (tolerance > 0.0 ) {
				horizon = 2 + (int)(Math.log(tolerance) / Math.log(Math.abs(z)));
				horizon = (horizon < c.length) ? (horizon) : (c.length);
			}
			zn = zn * zn;
			for (int n=1; n<horizon-1; n++) {
				zn = zn / z;
				sum = sum + (z1 + zn) * c[n];
				z1 = z1 * z;
			}
			return (sum / (1.0 - Math.pow(z, 2 * c.length - 2)));
		}
		
	}
}
