//=====================================================================================
//
// Project: 
// Constant-time bilateral filter
// 
// Authors: 
// Daniel Sage, Kunal N. Chaudhury
// Biomedical Imaging Group (BIG)
// Ecole Polytechnique Federale de Lausanne (EPFL)
// Lausanne, Switzerland
//
// Date:
// 4 February 2011
//
// Reference:
// Kunal N. Chaudhury, Daniel Sage, and Michael Unser,
// Fast O(1) bilateral filtering using trigonometric range kernels
// IEEE Transactions on Image Processing, submitted Dec. 2010
//
// Conditions of use:
// You'll be free to use this software for research purposes, but you
// should not redistribute it without our consent. In addition, we 
// expect you to include a citation or acknowledgment whenever 
// you present or publish results that are based on it.
//
//=====================================================================================
package bilateralfiltergui;

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.WindowManager;
import ij.gui.GUI;
import ij.process.ByteProcessor;
import ij.process.ColorProcessor;
import ij.process.FloatProcessor;
import ij.process.ShortProcessor;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import additionaluserinterface.GridPanel;
import additionaluserinterface.Settings;
import additionaluserinterface.SpinnerDouble;
import additionaluserinterface.SpinnerInteger;
import additionaluserinterface.WalkBar;
import bilateralfilterinstant.Data;
import bilateralfilterinstant.Filter;

public class MainDialog  extends JDialog implements WindowListener, ActionListener, ChangeListener, Runnable {

	private Settings settings = new Settings("Bilateral-Filter-Instant", IJ.getDirectory("plugins") + "Bilateral-Filter-Instant.txt");
	private WalkBar walk = new WalkBar("(c) 2011 EPFL, BIG", true, false, true);
	private Thread thread = null;
	
	private JButton 		bnRun 				= new JButton("Run");
	private SpinnerDouble 	spnRange 			= new SpinnerDouble(80, 1, 10000.0, 1);
	private SpinnerDouble 	spnSpace 			= new SpinnerDouble(20, 0, 10000.0, 1);
	private SpinnerDouble 	spnAnisotropyXY 	= new SpinnerDouble(1, 0, 10000.0, 0.1);
	private SpinnerDouble 	spnAnisotropyXZ 	= new SpinnerDouble(1, 0, 10000.0, 0.1);
	private SpinnerInteger 	spnNbIteration 		= new SpinnerInteger(1, 1, 1000, 1);
	private SpinnerDouble 	spnSelectChannels 	= new SpinnerDouble(0, 0, 100, 0.1);
	private JComboBox 		cmbModeIteration 	= new JComboBox(new String[] {"None", "Range", "Space", "Overvall"});
	private JLabel 			lblRMSE 			= new JLabel("RMSE");
	private JCheckBox		chkLog 				= new JCheckBox("Log", true);
	private JCheckBox 		chkMultithread 		= new JCheckBox("Multithread", true);
	private JComboBox 		cmbInput 			= new JComboBox();
	private JLabel 			lblRange 			= new JLabel("Dynamic range     ");
	private JLabel 			lblChannels			= new JLabel("Number of channels");
	private JLabel 			lblOrder 			= new JLabel("Order of the filter");
	private JLabel			lblIterations		= new JLabel("Iterations");
	private FilterCanvas 	canvas 				= new FilterCanvas();
	private JComboBox		cmbSelectChannels 	= new JComboBox(new String[] {"Best channels", "All channels", "By truncation", "By tolerance"}); 
	private JLabel 			lblValue 			= new JLabel("Value");

	private ImagePlus imp;
	private Data data;
	private Vector<ImagePlus> list;
	
	/**
	 * Constructor. 
	 * Get a reference on the current image. 
	 */
	public MainDialog() {
		super(new JFrame(), "Bilateral Filter Instant");
		imp = WindowManager.getCurrentImage();
		refreshInputList();
	}
	
	/**
	 * Build and show the main dialog box.
	 */
	public void doDialog() {
		walk.fillAbout("Bilateral Filter Instant", 
				"Version 04.02.2011", 
				"Implementation of the bilateral filter in constant-time using the raised cosines",
				"Daniel Sage & Kunal N. Chaudhury", 
				"Biomedical Imaging Group (BIG)<br>" + 
				"Ecole Polytechnique Federale de Lausanne (EPFL)<br>" +
				"Lausanne, Switzerland", 
				"<b>Reference:</b> " +
				"Kunal N. Chaudhury, Daniel Sage, and Michael Unser " +
				"'Fast O(1) bilateral filtering using trigonometric range kernels,' "+
				"IEEE Transactions on Image Processing, submitted Dec. 2010<br>", 
				"http://bigwww.epfl.ch/");
		
		settings.record("spnRange",				spnRange, "40");
		settings.record("spnSpace",				spnSpace, "80");
		settings.record("spnAnisotropyXY",		spnRange, "40");
		settings.record("spnAnisotropyXZ",		spnRange, "40");
		settings.record("chkLog", 				chkLog, true);
		settings.record("chkMultithread", 		chkMultithread, true);
		settings.record("spnNbIteration", 		spnNbIteration, "1");
		settings.record("cmbModeIteration", 	cmbModeIteration, "None");
		settings.record("spnSelectChannels",	spnSelectChannels, "0");
		settings.record("cmbSelectChannels",	cmbSelectChannels, "Best channels");
		settings.loadRecordedItems();
	
		lblChannels.setBorder(new CompoundBorder (BorderFactory.createEtchedBorder(),new EmptyBorder(2,5,2,2)));
		lblRange.setBorder(new CompoundBorder (BorderFactory.createEtchedBorder(),new EmptyBorder(2,5,2,2)));
		lblOrder.setBorder(new CompoundBorder (BorderFactory.createEtchedBorder(),new EmptyBorder(2,5,2,2)));
		lblRMSE.setBorder(new CompoundBorder (BorderFactory.createEtchedBorder(),new EmptyBorder(2,5,2,2)));
		canvas.setPreferredSize(new Dimension(200, 200));

			
		JPanel pnCanvas = new JPanel();
		pnCanvas.add(canvas);
		
		GridPanel params = new GridPanel("Standard", 5);
		params.place(1, 0, new JLabel("Sigma Range"));
		params.place(1, 1, 2, 1, spnRange);
		params.place(2, 0, new JLabel("Sigma Space"));
		params.place(2, 1, 2, 1, spnSpace);
		
		GridPanel advanced = new GridPanel("Advanced", 5);
		advanced.place(3, 0, new JLabel("Ratio X/Y"));
		advanced.place(3, 1, 2, 1, spnAnisotropyXY);
		advanced.place(4, 0, new JLabel("Ratio X/Z"));
		advanced.place(4, 1, 2, 1, spnAnisotropyXZ);
		advanced.place(5, 0, new JLabel("Iterate"));
		advanced.place(5, 1, cmbModeIteration);
		advanced.place(6, 0, lblIterations);
		advanced.place(6, 1, spnNbIteration);

		GridPanel tab0 = new GridPanel(5);
		tab0.place(0, 0, params);
		tab0.place(1, 0, advanced);
		
		GridPanel approx = new GridPanel(true, 5);
		approx.place(1, 0, new JLabel("Method"));
		approx.place(1, 1, 2, 1, cmbSelectChannels);
		approx.place(3, 0, lblValue);
		approx.place(3, 1, spnSelectChannels);
		approx.place(3, 2, new JLabel("%"));
		approx.place(4, 0, new JLabel("Order"));
		approx.place(4, 1, lblOrder);
		approx.place(4, 2, new JLabel("Terms"));
		approx.place(5, 0, new JLabel("Error"));
		approx.place(5, 1, lblRMSE);
		approx.place(5, 2, new JLabel("(RMSE)"));
		approx.place(8, 0, 3, 1, pnCanvas);
		
		GridPanel button = new GridPanel(true);
		//button.place(0, 0, 3, 1, chkPreview);
		button.place(1, 1, chkLog);
		button.place(1, 0, chkMultithread);
		button.place(2, 0, lblChannels);
		button.place(2, 1, bnRun);
	
		bnRun.addActionListener(this);
		
		JTabbedPane tab = new JTabbedPane();
		tab.add(tab0, "Parameters");
		tab.add(approx, "Approx.");
		
		GridPanel pn = new GridPanel();
		pn.place(0, 0, cmbInput);
		pn.place(1, 0, lblRange);
		pn.place(3, 0, tab);
		pn.place(4, 0, button);
		pn.place(5, 0, walk);
		
		walk.getButtonClose().addActionListener(this);
		
		spnSpace.addChangeListener(this);
		spnRange.addChangeListener(this);
		spnAnisotropyXY.addChangeListener(this);
		spnAnisotropyXZ.addChangeListener(this);
		cmbModeIteration.addActionListener(this);
		cmbInput.addActionListener(this);
		spnSelectChannels.addChangeListener(this);
		cmbSelectChannels.addActionListener(this);
		
		addWindowListener(this);
		getContentPane().add(pn);

		pack();
		setResizable(true);
		GUI.center(this);
		updateInterface();
		setVisible(true);
	}
	
	/**
	 * Implements the run for the Runnable.
	 */
	public synchronized  void actionPerformed(ActionEvent e) {
		Object source = e.getSource();
    	if (source == walk.getButtonClose()) {
			settings.storeRecordedItems();
     		dispose();
    	}
     	else if (source == cmbInput) {
    		selectInputImage();
    	}
     	else if (source == cmbSelectChannels) {
    		updateInterface();
    	}
     	else if (source == cmbModeIteration) {
    		updateInterface();
    	}
		else if (source == bnRun) {
			if (bnRun.getText().equals("Abort")) {
				if (thread != null) {
					thread.stop();
					thread = null;
					walk.finish("Aborted");
				}
				bnRun.setText("Run");
			}
			else {
				if (thread == null) {
					thread = new Thread(this);
					thread.setPriority(Thread.MIN_PRIORITY);
					thread.start();
				}
			}
		}
	}
		
	/**
	 * Implements the run for the Runnable.
	 */
	public void run() {
		walk.reset();
		bnRun.setText("Abort");
		float sigmaX = (float)spnSpace.get();
		float sigmaY = (float)sigmaX * (float)spnAnisotropyXY.get();
		float sigmaZ = (float)sigmaX * (float)spnAnisotropyXZ.get();
		float sigRange = (float)spnRange.get();
		int methodSelectionChannels = cmbSelectChannels.getSelectedIndex();
		double valueSelectionChannels = spnSelectChannels.get();
		int nbIteration = spnNbIteration.get();
		int modeIteration = cmbModeIteration.getSelectedIndex();
		boolean multithread = chkMultithread.isSelected();
		boolean log = chkLog.isSelected();
		process(sigRange, sigmaX, sigmaY, sigmaZ, methodSelectionChannels, valueSelectionChannels, 
				modeIteration, nbIteration, multithread, log);
		bnRun.setText("Run");
		walk.finish("Elapsed time: ");
		thread = null;
	}
	
	/**
	* Main routine to process the image.
	*/
	public void process(float sigRange, float sigmaX, float sigmaY, float sigmaZ, int methodSelectionChannels, 
			double valueSelectionChannels, int modeIteration, int nbIteration, boolean multithread, boolean log) {
		
		double chronoTotal = System.currentTimeMillis();		
			
		if (log) {
			IJ.log("------------------------------------------------------------");
			IJ.log("Bilateral filter on " + imp.getTitle());
			IJ.log("Parameter Sigma-Range: " + sigRange);
			IJ.log("Parameter Sigma-X/Y/Z: " + sigmaX + ", " + sigmaY + ", " + sigmaZ);
			IJ.log("Parameter Iteration: " + cmbModeIteration.getSelectedItem() + (modeIteration != 0 ? " number: " + nbIteration : ""));
			IJ.log("Parameter Selection: " + cmbSelectChannels.getSelectedItem() + (methodSelectionChannels > 2 ? " value: " + valueSelectionChannels : ""));
		}
		
		// --------------------------------------------------------------------------
		// Setting the filters
		// --------------------------------------------------------------------------
		Filter filter = new Filter(sigRange, sigmaX, sigmaY, sigmaZ);
		filter.setLog(log);
		filter.setMultithread(multithread);
		filter.setIterativeScheme(nbIteration, modeIteration);
		filter.selectChannels(methodSelectionChannels, valueSelectionChannels);
		int order[] = filter.computeOrder(data);
		int n = order[3]-order[2];
		if (n > 500)
			return;
		
		// --------------------------------------------------------------------------
		// Convert and process the input image
		// --------------------------------------------------------------------------
		int nx = imp.getWidth();
		int ny = imp.getHeight();
		int nz = imp.getStackSize();
		int nxy = nx*ny;

		ImageStack stack = new ImageStack(nx, ny);
		
		if (imp.getType() == ImagePlus.COLOR_RGB) {
			int rgb[][] = new int[nz][nxy];
			for(int c=0; c<3; c++) {
				float out[] = filter.execute(walk, data, c);
				int shift = 8*(2-c);
				for(int z=0; z<nz; z++)
				for(int k=0; k<nxy; k++) 
					rgb[z][k] += (byte)(out[z*nxy+k]) << shift;
			}
			for(int z=0; z<nz; z++) {
				ColorProcessor cp = new ColorProcessor(nx, ny);
				cp.setPixels(rgb[z]);
				stack.addSlice("", cp);
			}
		}
		else {
			float out[] = filter.execute(walk, data, 0);
			if (imp.getType() == ImagePlus.GRAY8) {
				for(int z=0; z<nz; z++) {
					byte gray[] = new byte[nxy];
					for(int x=0; x<nxy; x++)
						gray[x] = (byte)out[x + nxy*z];
					stack.addSlice("", new ByteProcessor(nx, ny, gray, null));
				}
			}
			else if (imp.getType() == ImagePlus.GRAY16) {
				for(int z=0; z<nz; z++) {
					short gray[] = new short[nxy];
					for(int x=0; x<nxy; x++)
						gray[x] = (short)out[x + nxy*z];
					stack.addSlice("", new ShortProcessor(nx, ny, gray, null));
				}
			}
			else if (imp.getType() == ImagePlus.GRAY32) {
				for(int z=0; z<nz; z++) {
					float gray[] = new float[nxy];
					for(int x=0; x<nxy; x++)
						gray[x] = out[x + nxy*z];
					stack.addSlice("", new FloatProcessor(nx, ny, gray, null));
				}
			}
		}
				
		if (stack.getSize() == imp.getStackSize()) {
			ImagePlus impout = new ImagePlus("BFI on " + imp.getTitle() + " (" + sigmaX + "-" + sigRange + ") ", stack);
			impout.show();
		}
		if (log)
			IJ.log("Total time: " +(System.currentTimeMillis()-chronoTotal));
	}
	
	/**
	 * Implements the stateChanged for the ChangeListener interface.
	 */
	public void stateChanged(ChangeEvent e) {
		updateInterface();
		if (thread != null) {
			thread.stop();
			thread = null;
		}
	}
	
	/**
	 * Update the status components of the dialog box.
	 */
	private void updateInterface() {
		
		spnSelectChannels.setEnabled(cmbSelectChannels.getSelectedIndex() > 1);
		lblValue.setEnabled(cmbSelectChannels.getSelectedIndex() > 1);
		//spnSelectChannels.setIncrement(cmbSelectChannels.getSelectedIndex() == 3 ? 0.1 : 1);
		float sigmaX = (float)spnSpace.get();
		float sigmaY = (float)sigmaX * (float)spnAnisotropyXY.get();
		float sigmaZ = (float)sigmaX * (float)spnAnisotropyXZ.get();
		float sigRange = (float)spnRange.get();

		if (cmbModeIteration.getSelectedIndex()==0) {
			spnNbIteration.set(1);
			spnNbIteration.setEnabled(false);
			lblIterations.setEnabled(false);
		}
		else {
			lblIterations.setEnabled(true);
			spnNbIteration.setEnabled(true);
		}
		
		Filter filter = new Filter(sigRange, sigmaX, sigmaY, sigmaZ);
		filter.selectChannels(cmbSelectChannels.getSelectedIndex(), spnSelectChannels.get());
		int order[] = filter.computeOrder(data);
		int n = order[3]-order[2];
		lblOrder.setText("" + n + " / " + order[1]);
		lblChannels.setText("" + n + " terms");
		
		if (n > 500) {
			canvas.setCurves(null, null, null);
			lblChannels.setText("Too many !");
		}
		int bins = 200;
		float gaussianApproximated[] = new float [bins];
		float gaussianDiscretized[] = new float[bins];
		float quality[] = filter.approximate(data, gaussianDiscretized, gaussianApproximated);
		canvas.setCurves(gaussianDiscretized, gaussianApproximated, quality);
		lblRMSE.setText(IJ.d2s(quality[0]*100, 1) + " %");
		
		if (n > 120)
			lblChannels.setBackground(Color.red);
		else if (n > 50)
			lblChannels.setBackground(Color.orange);
		else 
			lblChannels.setBackground(lblRange.getBackground());
		
		if (n > 500) {
			lblChannels.setBackground(Color.red);
			bnRun.setEnabled(false);
		}
		else
			bnRun.setEnabled(true);
	}

	/**
	 * Refresh the list of input images in the combo box.
	 */
	private void refreshInputList() {
		int[] ids = WindowManager.getIDList();
		list = new Vector<ImagePlus>();
		cmbInput.removeAllItems();
		int current = 0;
		boolean flag = false;
		if (ids != null)
		for(int i=0; i<ids.length; i++) {
			ImagePlus impt = WindowManager.getImage(ids[i]);
			if (impt == imp) {
				current = i;
				flag = true;
			}
			if (impt != null) {
				int type = impt.getType(); 
				if (type == ImagePlus.COLOR_RGB || type == ImagePlus.GRAY32 || type == ImagePlus.GRAY16 || type == ImagePlus.GRAY8) {
					list.add(impt);
				}
			}
		}
		
		for(int i=0; i<list.size(); i++) {
			ImagePlus impt = list.get(i);
			String title = impt.getTitle();
			if (title.length() > 28) 
				title = impt.getTitle().substring(0, 28) + "...";
			cmbInput.addItem(title);
		}
			
		
		if (flag) {
			cmbInput.setSelectedIndex(current);
		}

		if (list.size() == 0) {
			lblRange.setForeground(Color.red);
			lblRange.setText("Only 8, 16, 32, or RGB image");
			bnRun.setEnabled(false);
		}
		else {
			lblRange.setForeground(lblOrder.getForeground());
			bnRun.setEnabled(true);
			selectInputImage();
		}
	}
	
	
	/**
	 * Select  the input image.
	 */
	private void selectInputImage() {
		int index = cmbInput.getSelectedIndex();
		if (index < 0 || index >= list.size() )
			return;
		imp = list.get(index);
		if (imp == null)
			return;
		data = new Data(imp);
		bnRun.setEnabled(true);
		lblRange.setText("Input range: " + IJ.d2s(data.getMinimum()) + "..." + IJ.d2s(data.getMaximum()));
		updateInterface();
	}
	
	/**
	 * Implements the methods for the WindowListener.
	 */
	public void windowActivated(WindowEvent e) 		{ refreshInputList(); }
	public void windowClosed(WindowEvent e) 		{}
	public void windowDeiconified(WindowEvent e)	{}
	public void windowIconified(WindowEvent e)		{}
	public void windowOpened(WindowEvent e)			{}			
	public void windowClosing(WindowEvent e) 		{ dispose(); }
	public void windowDeactivated(WindowEvent e) 	{}
}
